//! find duplicate files
//!
//! Find duplicate files in a FileTree comparing their hash.

use crate::FileTree;
use std::collections::{HashMap, HashSet};
use std::fs;
use std::path::{Path, PathBuf};

// hash
use sha2::{Digest, Sha256};
use std::io::{BufReader, Read};

// parallelism
use rayon::prelude::*;
use std::sync::{Arc, Mutex};

impl FileTree {
    /// Find duplicates in a directory (including sub-directories).
    ///
    /// If path exist in tree, find duplicates using sha256. If 2 (or more) files have the same hash they're duplicates.
    ///
    /// returns a `Vec` containing a `Vec<PathBuf>` for each group of duplicates.
    ///
    /// If path doesn't exist, or if no duplicates are found, return None
    ///
    /// # Examples
    /// Assuming a directory structure like this. files in lowercase, directories in uppercase. And assuming same letter means duplicate. dups will contain a Some(Vec<Vec<_>>) with a,A/a,B/a
    ///
    /// ./{A/a,B/a,a,C/c}
    ///
    /// ```no_run
    /// let p = Path::new(".");
    /// let f = FileTree::new(&p, &false);
    /// let dups = f.duplicates(&p);
    /// ```
    pub fn duplicates(&self, path: &Path) -> Option<Vec<Vec<PathBuf>>> {
        match self.get_children(path) {
            Some(value) => {
                // every path in subdir (directories+files)
                let children: HashSet<PathBuf> = value.into_iter().collect();
                // every files (whole tree)
                let files: HashSet<PathBuf> = self.files().into_iter().collect();

                // every files in subdir
                let intersection: Vec<_> = children.intersection(&files).cloned().collect();

                // Arc<Mutex<_>> used to safely update HashMap in parallel
                let hashes: Arc<Mutex<HashMap<[u8; 32], Vec<PathBuf>>>> =
                    Arc::new(Mutex::new(HashMap::new()));

                // parallelized loop
                intersection.par_iter().for_each(|item| {
                    if let Ok(file) = fs::File::open(item) {
                        let mut sha256 = Sha256::new();

                        // sizable buffer
                        let mut buffer = [0; 8192];
                        let mut reader = BufReader::new(file);

                        // https://rust-lang-nursery.github.io/rust-cookbook/cryptography/hashing.html
                        loop {
                            let count = reader.read(&mut buffer).unwrap();
                            // rust complains with .ok()?
                            // unwrap should not fail in this context though
                            if count == 0 {
                                break;
                            }
                            sha256.update(&buffer[..count]);
                        }

                        let hash = sha256.finalize();

                        // Use a Mutex to update HashMap in parallel
                        let mut locked_hashes = hashes.lock().unwrap();
                        locked_hashes
                            .entry(hash.into())
                            .or_default()
                            .push(item.clone());
                    }
                });

                // extract result from Mutex
                // converting Arc<Mutex<HashMap>> into HashMap
                let hashes = Arc::try_unwrap(hashes).ok().unwrap().into_inner().unwrap();

                let dups = hashes
                    .values()
                    .filter(|a| a.len() > 1) // if more than 1 path for a hash it's a duplicate
                    .cloned()
                    .collect::<Vec<Vec<PathBuf>>>();

                if dups.is_empty() {
                    None
                } else {
                    Some(dups)
                }
            }
            _ => None,
        }
    }

    /// Collision shouldn't happen with sha256, but a method to check if there is a collision in duplicates found
    /// # Examples
    /// ```no_run
    /// let p = Path::new(".");
    /// let f = FileTree::new(&p, &false);
    /// let dups = f.duplicates(&p);
    /// let check = FileTree::collision(dups);
    /// ```
    pub fn collision(dups: Option<Vec<Vec<PathBuf>>>) -> bool {
        match dups {
            Some(value) => {
                for mut subgroup in value {
                    let path1 = subgroup.pop().unwrap();
                    while let Some(path2) = subgroup.pop() {
                        if !Self::raw_compare(&path1, &path2) {
                            return true; // files contents are different, it's a collision
                        }
                    }
                }
                false
            }
            None => false, // no collision since there is any duplicate
        }
    } // no unittest for collision as finding a collision is highly unlikely, and there are no known collision yet

    /// Check if 2 files have the same content character to character, return a boolean.
    fn raw_compare(path1: &Path, path2: &Path) -> bool {
        if let Ok(file1) = fs::File::open(path1) {
            if let Ok(file2) = fs::File::open(path2) {
                // sizable buffer
                let mut buffer1 = [0; 8192];
                let mut buffer2 = [0; 8192];
                let mut reader1 = BufReader::new(file1);
                let mut reader2 = BufReader::new(file2);
                loop {
                    let count1 = reader1.read(&mut buffer1).unwrap();
                    let count2 = reader2.read(&mut buffer2).unwrap();
                    if count1 == 0 || count2 == 0 {
                        if count1 != count2 {
                            return false;
                        } // files do not have same size
                        break;
                    }
                    if buffer1[..count1] != buffer2[..count2] {
                        return false;
                    }
                }
                true
            } else {
                panic!("{}", format!("couldn't read file {}", path2.display()))
            }
        } else {
            panic!("{}", format!("couldn't read file {}", path1.display()))
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use std::fs::File;
    use std::io::prelude::*;
    use std::process::Command;
    #[test]
    fn duplicate_test() {
        // create temporary test directories and files
        let _ = Command::new("mkdir")
            .arg("/tmp/test_du_dups1/dir1/dir11")
            .arg("-p")
            .spawn()
            .unwrap()
            .wait();
        let _ = Command::new("mkdir")
            .arg("/tmp/test_du_dups1/dir1/dir12")
            .arg("-p")
            .spawn()
            .unwrap()
            .wait();

        // let's create the same file 3 times
        let mut file = File::create("/tmp/test_du_dups1/file1").unwrap();
        let _ = file.write_all(b"Hello, world!").unwrap();
        let _ = Command::new("cp")
            .arg("/tmp/test_du_dups1/file1")
            .arg("/tmp/test_du_dups1/dir1/")
            .arg("-p")
            .spawn()
            .unwrap()
            .wait();
        let _ = Command::new("cp")
            .arg("/tmp/test_du_dups1/file1")
            .arg("/tmp/test_du_dups1/file2")
            .arg("-p")
            .spawn()
            .unwrap()
            .wait();
        let _ = Command::new("cp")
            .arg("/tmp/test_du_dups1/file1")
            .arg("/tmp/test_du_dups1/dir1/dir11/")
            .arg("-p")
            .spawn()
            .unwrap()
            .wait();
        // let's create a different set of duplicates
        let mut file1 = File::create("/tmp/test_du_dups1/file10").unwrap();
        let _ = file1.write_all(b"lorem ipsum dolor sit amet").unwrap();
        let _ = Command::new("cp")
            .arg("/tmp/test_du_dups1/file10")
            .arg("/tmp/test_du_dups1/dir1/")
            .arg("-p")
            .spawn()
            .unwrap()
            .wait();
        // and a unique file here
        let mut file2 = File::create("/tmp/test_du_dups1/file100").unwrap();
        let _ = file2.write_all(b"consectetur adipiscing elit").unwrap();

        let p = Path::new("/tmp/test_du_dups1");
        let f = FileTree::new(&p, &false).unwrap();
        let mut res = f.duplicates(&p).unwrap(); // unwrap asserts it's not none

        res.sort_by_key(|b| b.len());

        let expected_vec: Vec<Vec<PathBuf>> = vec![
            vec![
                Path::new("/tmp/test_du_dups1/file10").to_path_buf(),
                Path::new("/tmp/test_du_dups1/dir1/file10").to_path_buf(),
            ],
            vec![
                Path::new("/tmp/test_du_dups1/file1").to_path_buf(),
                Path::new("/tmp/test_du_dups1/dir1/dir11/file1").to_path_buf(),
                Path::new("/tmp/test_du_dups1/dir1/file1").to_path_buf(),
                Path::new("/tmp/test_du_dups1/file2").to_path_buf(),
            ],
        ];

        // using a HashMap to find duplicates, items are not ordered, so we compare them using a set
        // first set of duplicates
        let expected = expected_vec[0].iter().cloned().collect::<HashSet<_>>();
        let actual = res[0].iter().cloned().collect::<HashSet<_>>();
        assert_eq!(expected, actual);
        // second set of duplicates
        let expected = expected_vec[1].iter().cloned().collect::<HashSet<_>>();
        let actual = res[1].iter().cloned().collect::<HashSet<_>>();
        assert_eq!(expected, actual);

        // sets do not allow multiple elements so we have to check vec length separately
        assert_eq!(expected_vec[0].len(), res[0].len());
        assert_eq!(expected_vec[1].len(), res[1].len());

        // we check if we have same number of duplicate groups
        assert_eq!(expected_vec.len(), res.len());

        //clean
        let _ = Command::new("rm")
            .arg("-rf")
            .arg("/tmp/test_du_dups1")
            .spawn()
            .unwrap()
            .wait();
    }
    #[test]
    fn duplicate_test_empty() {
        // create temporary test directories and files
        let _ = Command::new("mkdir")
            .arg("/tmp/test_du_dups2/dir1")
            .arg("-p")
            .spawn()
            .unwrap()
            .wait();
        let mut file = File::create("/tmp/test_du_dups2/file10").unwrap();
        let _ = file
            .write_all(b"Integer faucibus sapien vitae aliquet dapibus")
            .unwrap();
        let mut file1 = File::create("/tmp/test_du_dups2/dir1/file10").unwrap();
        let _ = file1.write_all(b"Pellentesque at pharetra enim").unwrap();

        let p = Path::new("/tmp/test_du_dups2");
        let f = FileTree::new(&p, &false).unwrap();
        let res = f.duplicates(&p);
        // should not find duplicates
        assert!(res.is_none());

        // clean
        let _ = Command::new("rm")
            .arg("-rf")
            .arg("/tmp/test_du_dups2")
            .spawn()
            .unwrap()
            .wait();
    }

    #[test]
    fn file_compare() {
        // create temporary test directory and files
        let _ = Command::new("mkdir")
            .arg("/tmp/test_du_compare")
            .arg("-p")
            .spawn()
            .unwrap()
            .wait();
        let mut file = File::create("/tmp/test_du_compare/file1").unwrap();
        let _ = file.write_all(b"Cras eleifend nisi nibh").unwrap();
        let mut file = File::create("/tmp/test_du_compare/file2").unwrap();
        let _ = file.write_all(b"Cras eleifend nisi nibh").unwrap();
        let mut file = File::create("/tmp/test_du_compare/file3").unwrap();
        let _ = file.write_all(b"a blandit elit mattis et").unwrap();

        let p1 = Path::new("/tmp/test_du_compare/file1");
        let p2 = Path::new("/tmp/test_du_compare/file2");
        let p3 = Path::new("/tmp/test_du_compare/file3");

        assert!(FileTree::raw_compare(p1, p2));
        assert!(!FileTree::raw_compare(p1, p3));

        // clean
        let _ = Command::new("rm")
            .arg("-rf")
            .arg("/tmp/test_du_compare")
            .spawn()
            .unwrap()
            .wait();
    }
}
