//! tree structure representation
//!
//! Implement a tree structure to represent data on a disk (files and directories).
//!
//! Include methods to
//! + construct this tree
//! + get a directory's children
//! + get a file or directory size
//! + get files in tree.

use crate::size::Size;
use std::collections::HashMap;
use std::fs;
use std::io::{Error, ErrorKind};
use std::path::{Path, PathBuf};

/// FileTree structure
pub struct FileTree {
    /// tree root (base path)
    root: PathBuf,
    /// HashMap containing every path in the tree (every files and directories, sub-directories included) and their associated data (size and children).
    map: HashMap<PathBuf, EntryNode>,
}

enum EntryNode {
    File { size: Size },
    Path { children: Vec<PathBuf>, size: Size },
}

impl FileTree {
    /// Create a new filetree from given path.
    ///
    /// Will return an error if path doesn't exist or if user doesn't have read permission.
    ///
    /// If user is missing read permission in a subdir, will ignore this subdir and keep building the tree.
    ///
    /// access_denied=true show ignored paths, access_denied=false to hide that these paths were ignored.
    ///
    /// # Examples
    ///
    /// ```no_run
    /// let p = Path::new(".");
    /// let f = FileTree::new(&p, false);
    /// ```
    pub fn new(root: &Path, access_denied: &bool) -> std::io::Result<Self> {
        if root.exists() {
            if let Ok(meta) = fs::metadata(root) {
                let r = root.to_path_buf();

                if meta.is_dir() {
                    let entries = fs::read_dir(root)?;
                    let mut paths: Vec<PathBuf> = entries
                        .filter_map(|entry| match entry {
                            Ok(entry) => Some(entry.path()),
                            Err(_) => None,
                        })
                        .collect();
                    let e = EntryNode::Path {
                        children: paths.clone(),
                        size: Size::new(0),
                    };
                    let mut map_entry: Vec<(PathBuf, EntryNode)> = Vec::new();
                    map_entry.push((r.clone(), e));
                    while let Some(current_path) = paths.pop() {
                        if !(current_path.is_symlink()) {
                            // need to find a better way to handle symlinks
                            if let Ok(meta) = fs::metadata(current_path.clone()) {
                                if meta.is_dir() {
                                    if let Ok(entries) = fs::read_dir(&current_path) {
                                        let new_paths: Vec<PathBuf> = entries
                                            .filter_map(|entry| match entry {
                                                Ok(entry) => Some(entry.path()),
                                                Err(_) => None,
                                            })
                                            .collect();
                                        let e = EntryNode::Path {
                                            children: new_paths.clone(),
                                            size: Size::new(0),
                                        };
                                        map_entry.push((current_path.clone(), e));
                                        paths.extend(new_paths);
                                    } else if *access_denied {
                                        println!("{} children not processed, might be missing read permission",current_path.display());
                                    }
                                } else {
                                    // file
                                    let e = EntryNode::File {
                                        size: Size::new(meta.len()),
                                    };
                                    map_entry.push((current_path, e));
                                }
                            } else if *access_denied {
                                println!(
                                    "{} not processed, might be missing read permission",
                                    current_path.display()
                                );
                            }
                        }
                    }

                    let hashmap: HashMap<PathBuf, EntryNode> = map_entry.into_iter().collect();
                    let mut tree = Self {
                        root: r.clone(),
                        map: hashmap,
                    };
                    tree.set_size();
                    Ok(tree)
                } else {
                    let e = EntryNode::File {
                        size: Size::new(meta.len()),
                    };

                    Ok(Self {
                        root: r.clone(),
                        map: HashMap::from([(r, e)]),
                    })
                }
            } else {
                Err(Error::new(
                    ErrorKind::Other,
                    format!("Error processing path {}", root.display()),
                ))
            }
        } else {
            Err(Error::new(ErrorKind::Other, "path not found"))
        }
    }

    /// set attribute size of every EntryNode in the tree
    fn set_size(&mut self) {
        let mut paths: Vec<_> = self.map.keys().cloned().collect();
        paths.sort_by_key(|b| std::cmp::Reverse(b.as_os_str().len())); // path for a child is necessarilly longer than its parent's path so we process items with longest path first.
        for path in paths {
            let mut s = Size::new(0);
            if let Some(EntryNode::Path { children, .. }) = self.map.get_mut(path.as_path()) {
                for child in children.clone() {
                    // children were already processed before so we can get their size
                    if let Some(EntryNode::Path { size, .. }) | Some(EntryNode::File { size }) =
                        self.map.get(child.as_path())
                    {
                        s = s + *size;
                    }
                }
            }
            if let Some(EntryNode::Path { ref mut size, .. }) = self.map.get_mut(path.as_path()) {
                *size = s;
            }
        }
    }

    /// return FileTree root
    /// # Examples
    ///
    /// ```no_run
    /// let p = Path::new(".");
    /// let f = FileTree::new(&p, &false);
    /// let root = f.get_root();
    /// assert_eq!(root, p);
    /// ```
    pub fn get_root(&self) -> &Path {
        self.root.as_path()
    }

    /// return a Vec with every paths in the subtree found from given path (recursively)
    ///
    /// return None if given path doesn't exist in the FileTree
    /// # Examples
    /// assuming a directories structure like this, children will contains a (a eventually contains files and directories in sub-directories).
    ///
    /// ./{A/a,B}
    /// ```no_run
    /// let p = Path::new(".");
    /// let f = FileTree::new(&p, &false);
    /// let children = f.get_children(&Path::new("./A"));
    /// ```
    pub fn get_children(&self, path: &Path) -> Option<Vec<PathBuf>> {
        if let Some(vec_paths) = self.get_direct_children(path) {
            let mut paths = vec_paths;
            let mut temp_paths = paths.clone();
            while let Some(current_path) = temp_paths.pop() {
                if let Some(vec_paths) = self.get_direct_children(current_path.as_path()) {
                    paths.extend(vec_paths.clone());
                    temp_paths.extend(vec_paths);
                }
            }
            Some(paths)
        } else {
            None
        }
    }

    /// return direct children of a given path (non recursively, depth=1)
    ///
    /// None if path is a file, or doesn't exist.
    ///
    /// A `Vec<PathBuf>` if it's a directory.
    /// # Examples
    /// assuming a directories structure like this, children will contains a and b, but not a/c.
    ///
    /// ./{A/{a/c,b},B}
    /// ```no_run
    /// let p = Path::new(".");
    /// let f = FileTree::new(&p, &false);
    /// let children = f.get_children(&Path::new("./A"));
    /// ```
    pub fn get_direct_children(&self, path: &Path) -> Option<Vec<PathBuf>> {
        match self.map.get(path) {
            Some(EntryNode::Path { children, .. }) => Some(children.clone()),
            Some(EntryNode::File { .. }) => None,
            _ => None,
        }
    }

    /// If given path exists in the FileTree, return its size. None if it doesn't. Size can eventually be 0, an empty directory for example.
    /// # Examples
    /// size of entire tree (size from root):
    /// ```no_run
    /// let p = Path::new(".");
    /// let f = FileTree::new(&p);
    /// let size = f.get_size(&p);
    /// ```
    pub fn get_size(&self, path: &Path) -> Option<Size> {
        match self.map.get(path) {
            Some(EntryNode::File { size }) | Some(EntryNode::Path { size, .. }) => Some(*size),
            _ => None,
        }
    }

    /// return a Vec with all the files in FileTree (and only the files), can be empty.
    /// # Examples
    /// assuming a directories structure like this, lowercase for files, uppercase for directories. files will only contain a, b and c.
    ///
    /// ./{A/a,B/b,c}
    ///
    /// ```no_run
    /// let p = Path::new(".");
    /// let f = FileTree::new(&p, &false);
    /// let files = f.get_files();
    /// ```
    // an iterator instead of a Vec<PathBuf> would be lazy,
    // however, we use files to find an intersection with children in a subtree to find duplicates, an iterator wouldn't be useful for this
    pub fn files(&self) -> Vec<PathBuf> {
        let mut files = self
            .map
            .iter()
            .filter(|(key, _value)| !key.is_dir())
            .map(|(key, _value)| key.clone())
            .collect::<Vec<PathBuf>>();
        // we order by name because original hashmap is not ordered
        files.sort_by(|a, b| a.file_name().cmp(&b.file_name()));
        files
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use std::fs::File;
    use std::io::prelude::*;
    use std::process::Command;
    fn set_temp(number: u8) {
        // create temporary test directories and files
        let _ = Command::new("mkdir")
            .arg(format!("/tmp/test_du{}/dir1", number))
            .arg("-p")
            .spawn()
            .unwrap()
            .wait();
        let mut file = File::create(format!("/tmp/test_du{}/file1", number)).unwrap();
        let _ = file.write_all(b"Hello, world!").unwrap();
    }
    fn clear_temp(number: u8) {
        let _ = Command::new("rm")
            .arg("-rf")
            .arg(format!("/tmp/test_du{}", number))
            .spawn()
            .unwrap()
            .wait();
    }

    #[test]
    fn simple_filetree() {
        set_temp(1);

        let p = Path::new("/tmp/test_du1");
        let p_buf = p.to_path_buf();
        let f = FileTree::new(&p, &false).unwrap();
        let paths = [
            Path::new("/tmp/test_du1").to_path_buf(),
            Path::new("/tmp/test_du1/dir1").to_path_buf(),
            Path::new("/tmp/test_du1/file1").to_path_buf(),
        ];

        assert_eq!(f.root, p_buf);
        for path in paths {
            assert_eq!(f.map.contains_key(&path), true);
        }

        // for root
        let expected_children = vec![
            Path::new("/tmp/test_du1/dir1").to_path_buf(),
            Path::new("/tmp/test_du1/file1").to_path_buf(),
        ];
        let expected_size = "13"; // "Hello, wold!" has 13 characters, so file1 = 26B, directories are 0B
        let actual_value = f.map.get(&p_buf).unwrap();
        match actual_value {
            EntryNode::Path {
                children: actual_children,
                size,
            } => {
                assert_eq!(*actual_children, expected_children);
                assert_eq!(format!("{}", *size), expected_size)
            }
            _ => panic!(), // shouldn't happen this entry refers to a dir
        }
        // for dir1
        let expected_children: Vec<PathBuf> = vec![];
        let actual_value = f
            .map
            .get(&Path::new("/tmp/test_du1/dir1").to_path_buf())
            .unwrap();
        match actual_value {
            EntryNode::Path {
                children: actual_children,
                size,
            } => {
                assert_eq!(*actual_children, expected_children);
                assert_eq!(format!("{}", *size), "0"); // dir1 is empty
            }
            _ => panic!(), // shouldn't happen this entry refers to a dir
        }
        // for file1
        let expected_size = "13";
        let actual_value = f
            .map
            .get(&Path::new("/tmp/test_du1/file1").to_path_buf())
            .unwrap();
        match actual_value {
            EntryNode::File { size } => assert_eq!(format!("{}", *size), expected_size),
            _ => panic!(), // this entry refers to a file
        }
        clear_temp(1);
    }
    #[test]
    fn simple_filetree_file() {
        set_temp(2);
        // for a file
        let p2 = Path::new("/tmp/test_du2/file1");
        let p_buf2 = p2.to_path_buf();
        let f2 = FileTree::new(&p2, &false).unwrap();

        assert_eq!(f2.root, p_buf2);
        assert_eq!(f2.map.contains_key(&p_buf2), true);

        let expected_size = "13"; // "Hello, world!" has 13 characters
        let actual_value = f2.map.get(&p_buf2).unwrap();
        match actual_value {
            EntryNode::File { size } => assert_eq!(format!("{}", *size), expected_size),
            _ => panic!(), // shouldn't happen f2 refers to a file
        }

        // clean
        clear_temp(2);
    }
    #[test]
    fn filetree_fail() {
        let p = Path::new("/tmp/test_du/shouldntexist");
        let f = FileTree::new(&p, &false);
        assert!(f.is_err());
    }

    #[test]
    fn get_root_test() {
        set_temp(3);
        let p = Path::new("/tmp/test_du3");
        let f = FileTree::new(&p, &false).unwrap();
        assert_eq!(f.get_root(), p);
        clear_temp(3);
    }

    #[test]
    fn files_test() {
        set_temp(4);
        let mut file = File::create("/tmp/test_du4/dir1/file2").unwrap();
        let _ = file.write_all(b"Hello, world!").unwrap();
        let p = Path::new("/tmp/test_du4");
        let f = FileTree::new(&p, &false).unwrap();
        let expected = vec![
            Path::new("/tmp/test_du4/file1").to_path_buf(),
            Path::new("/tmp/test_du4/dir1/file2").to_path_buf(),
        ];
        assert_eq!(f.files(), expected);
        clear_temp(4);
    }

    #[test]
    fn children() {
        set_temp(5);
        let mut file = File::create("/tmp/test_du5/dir1/file2").unwrap();
        let _ = file.write_all(b"Hello, world!").unwrap();
        let p = Path::new("/tmp/test_du5");
        let f = FileTree::new(&p, &false).unwrap();
        let expected = vec![
            Path::new("/tmp/test_du5/dir1").to_path_buf(),
            Path::new("/tmp/test_du5/file1").to_path_buf(),
            Path::new("/tmp/test_du5/dir1/file2").to_path_buf(),
        ];
        assert_eq!(f.get_children(&p), Some(expected));

        let p = Path::new("/tmp/test_du5/dir1");
        let expected = vec![Path::new("/tmp/test_du5/dir1/file2").to_path_buf()];
        assert_eq!(f.get_children(&p), Some(expected));

        let p = Path::new("/tmp/test_du5/file1");
        assert_eq!(f.get_children(&p), None);

        let p = Path::new("/tmp/test_du5/shouldntexist");
        assert_eq!(f.get_children(&p), None);

        clear_temp(5);
    }

    #[test]
    fn size_test() {
        set_temp(6);
        let mut file = File::create("/tmp/test_du6/dir1/file2").unwrap();
        let _ = file.write_all(b"Hello, world!").unwrap();
        let _ = Command::new("mkdir")
            .arg("/tmp/test_du6/dir3")
            .arg("-p")
            .spawn()
            .unwrap()
            .wait();
        let p = Path::new("/tmp/test_du6");
        let f = FileTree::new(&p, &false).unwrap();

        let expected_size = "26"; // file1 and file2
        assert_eq!(format!("{}", f.get_size(&p).unwrap()), expected_size);

        let p = Path::new("/tmp/test_du6/file1");
        let expected_size = "13";
        assert_eq!(format!("{}", f.get_size(&p).unwrap()), expected_size);

        let p = Path::new("/tmp/test_du6/dir3");
        let expected_size = "0";
        assert_eq!(format!("{}", f.get_size(&p).unwrap()), expected_size);

        let p = Path::new("/tmp/test_du6/shouldntexist");
        assert!(f.get_size(&p).is_none());

        clear_temp(6);
    }
}
