//! Basic disk usage analyzer
//!
//! Include a command to show disk usage and file tree, ordered by size (default) or name. It is possible to filter files by "end with".
//!
//! Include a command to find duplicate files.
//! # Examples
//! show ~ disk usage
//! ```
//! $ cargo run -- usage ~
//! ```
//! show pdf files in ~/Downloads , ordered by name
//! ```
//! $ cargo run -- usage ~/Downloads --sort name --filter ".pdf"
//! ```
//! same command but with short options
//! ```
//! $ cargo run -- usage ~/Downloads -s name -f ".pdf"
//! ```
//! show duplicates in /tmp
//! ```
//! $ cargo run -- duplicates /tmp
//! ```
//! show duplicates in /tmp and check for collision
//! ```
//! $ cargo run -- duplicates /tmp --check
//! ```
//! same command but with short option
//! ```
//! $ cargo run -- duplicates /tmp -c
//! ```
//! for both duplicates and usage, to show directories not processed because of missing read permission
//! ```
//! $ cargo run -- duplicates ~ --access-denied
//! $ cargo run -- usage ~ --access-denied
//! ```
//! same with short option
//! ```
//! $ cargo run -- duplicates ~ -a
//! $ cargo run -- usage ~ -a
//! ```

mod duplicates;
mod file_tree;
mod print_tree;
mod size;

use clap::{Parser, Subcommand};
use file_tree::FileTree;
use std::path::{Path, PathBuf};

#[derive(Parser)]
#[command(author, version, about, long_about = None)]
#[command(propagate_version = true)]
struct Cli {
    #[command(subcommand)]
    command: Commands,
}

#[derive(Subcommand)]
enum Commands {
    /// Show the disk usage tree for the given path
    Usage {
        /// (default '.')
        path: Option<PathBuf>,
        /// (default "size")
        #[arg(long = "sort", short = 's')]
        sort: Option<String>,
        /// (default None)
        #[arg(long = "filter", short = 'f')]
        filter: Option<String>,
        /// (default false)
        #[arg(long = "access-denied", short = 'a')]
        access: bool,
    },
    /// Find duplicates in given path
    Duplicates {
        /// (default '.')
        path: Option<PathBuf>,
        /// (default false)
        #[arg(long = "check", short = 'c')]
        check: bool,
        /// (default false)
        #[arg(long = "access-denied", short = 'a')]
        access: bool,
    },
}

fn main() -> std::io::Result<()> {
    let cli = Cli::parse();

    match &cli.command {
        Commands::Usage {
            path,
            sort,
            filter,
            access,
        } => {
            let path = path.as_deref().unwrap_or(Path::new("."));
            match sort {
                Some(ref s) if s == "size" => {
                    FileTree::new(path, access)?.show(0, filter.as_deref())
                }
                None => FileTree::new(path, access)?.show(0, filter.as_deref()),
                Some(ref s) if s == "name" => {
                    FileTree::new(path, access)?.show(1, filter.as_deref())
                }
                _ => (),
            }
        }
        Commands::Duplicates {
            path,
            check,
            access,
        } => {
            let path = path.as_deref().unwrap_or(Path::new("."));
            let duplicate_files = FileTree::new(path, access)?.duplicates(path);
            if let Some(ref item_vec) = duplicate_files {
                for (i, item) in item_vec.iter().enumerate() {
                    println!("DUPLICATES {}", i + 1);
                    for i in item {
                        println!("{}", i.display());
                    }
                    println!("\n")
                }
            } else {
                println!("No duplicate found in {}\n", path.display());
            }
            if *check {
                println!("collision found: {}", FileTree::collision(duplicate_files));
            }
        }
    }
    Ok(())
}
