//! Print human readable file tree.
//!
//! Print nodes in a filetree (files and directories) with their sizes and tree structure.
//!
//! Include filter and sort arguments.
//!
//! Filter by "end with" (preserve the tree structure when printing).
//! Sort by name or by size.

use crate::file_tree::FileTree;
use crate::size::Size;
use std::path::Path;

impl FileTree {
    /// Print a FileTree in the console in a human readable format.
    ///
    /// sort = 0 is size order
    ///
    /// sort = 1 is lexicographical order
    ///
    /// Any other integer for sort won't fail but nodes won't be sorted and order might seem random.
    ///
    /// filter_suffix is for "end with" filter, will still preserve tree structure (will show parent directories if one descendant pass the filter).
    ///
    /// use filter_suffix = None to show everything
    /// # Examples
    ///
    /// ```no_run
    /// let p = Path::new(".");
    /// let f = FileTree::new(&p);
    /// f.show(1,None); // show everything ordered by name
    /// f.show(0,Some(".pdf")); // show pdf files (preserving tree structure) ordered by size
    /// ```
    pub fn show(&self, sort: u8, filter_suffix: Option<&str>) {
        let root = self.get_root();
        self.print_entry(root, &self.get_size(root).unwrap(), 0);
        self.show_recursive(root, 1, sort, filter_suffix);
    }

    /// recursively visit a FileTree to print it
    ///
    /// do not use alone, wrapper is show()
    fn show_recursive(&self, path: &Path, indent: usize, sort: u8, filter_suffix: Option<&str>) {
        if let Some(mut children) = self.get_direct_children(path) {
            match sort {
                0 => {
                    // Sort children by size in descending order
                    children.sort_by_key(|b| std::cmp::Reverse(self.get_size(b)))
                }
                1 => children.sort_by(|a, b| a.file_name().cmp(&b.file_name())),
                _ => (),
            }

            for child in children {
                if let Some(entry) = self.get_size(&child) {
                    if let Some(suffix) = filter_suffix {
                        let mut descendant = self.get_children(&child).unwrap_or_default();
                        descendant.push(child.clone());

                        let mut found = false;

                        // we check if one descendant pass the filter
                        // currently though we do it more than necessary
                        // might be more efficient to work with get_direct_children and memorize which parents pass the filter
                        // might be possible to visit descendant in reverse order and have a bit with true or false (default false) for each item
                        for item in descendant {
                            if item
                                .clone()
                                .into_os_string()
                                .into_string()
                                .unwrap()
                                .ends_with(suffix)
                            {
                                found = true;
                                break;
                            }
                        }

                        if found {
                            self.print_entry(&child, &entry, indent);
                        }
                    } else {
                        self.print_entry(&child, &entry, indent);
                    }

                    if let Some(_grandchildren) = self.get_direct_children(&child) {
                        self.show_recursive(&child, indent + 1, sort, filter_suffix);
                    }
                }
            }
        }
    }

    /// print a single entry
    ///
    /// do not use alone, is part os show_recursive()
    fn print_entry(&self, path: &Path, size: &Size, indent: usize) {
        let indentation = "\t".repeat(indent);
        println!("{}{}\t{}", indentation, size, path.display());
    }
}

// no unittest here for now as show is mostly "visual"
